@@ -23,6 +23,12 @@
 
 #ifdef NANONEXT_HTTP
 #include <nng/supplemental/http/http.h>
+#include "nng_structs.h"
+#ifdef _WIN32
+#include <winsock2.h>
+#else
+#include <arpa/inet.h>
+#endif
 
 typedef struct nano_handle_s {
   nng_url *url;
@@ -81,6 +87,9 @@ typedef struct nano_handle_s {
 #include <mbedtls/ctr_drbg.h>
 #include <mbedtls/md.h>
 #include <mbedtls/error.h>
+#ifdef MBEDTLS_PSA_CRYPTO_C
+#include <psa/crypto.h>
+#endif
 #include <errno.h>
 #endif
 
@@ -126,7 +135,7 @@ extern int R_interrupts_pending;
 #define NANONEXT_WAIT_DUR 1000
 #define NANONEXT_SLEEP_DUR 200
 #define NANO_ALLOC(x, sz)                                      \
-  (x)->buf = calloc(sz, sizeof(unsigned char));                \
+  (x)->buf = malloc(sz);                                       \
   if ((x)->buf == NULL) Rf_error("memory allocation failed");  \
   (x)->len = sz;                                               \
   (x)->cur = 0
@@ -141,6 +150,7 @@ extern int R_interrupts_pending;
   SET_STRING_ELT(klass, 0, Rf_mkChar(cls1));                   \
   SET_STRING_ELT(klass, 1, Rf_mkChar(cls2))
 #define NANO_ENSURE_ALLOC(x) if (x == NULL) { xc = 2; goto failmem; }
+#define NANO_URL_MAX 8192
 
 typedef union nano_opt_u {
   char *str;
@@ -159,6 +169,7 @@ typedef struct nano_stream_s {
   } endpoint;
   nng_tls_config *tls;
   int textframes;
+  int msgmode;
   enum {
     NANO_STREAM_DIALER,
     NANO_STREAM_LISTENER
@@ -230,7 +241,6 @@ typedef struct nano_serial_bundle_s {
   R_outpstream_t outpstream;
   R_inpstream_t inpstream;
   SEXP klass;
-  unsigned char *buf;
 } nano_serial_bundle;
 
 typedef enum nano_list_op {
@@ -241,10 +251,112 @@ typedef enum nano_list_op {
   SHUTDOWN
 } nano_list_op;
 
-typedef struct nano_node_s {
-  void *data;
-  struct nano_node_s *next;
-} nano_node;
+#ifdef NANONEXT_HTTP
+
+// Forward declarations
+typedef struct nano_conn_s nano_conn;
+typedef struct nano_ws_conn_s nano_ws_conn;
+typedef struct nano_stream_conn_s nano_stream_conn;
+typedef struct nano_http_server_s nano_http_server;
+typedef struct nano_http_handler_info_s nano_http_handler_info;
+typedef struct nano_http_request_s nano_http_request;
+
+// Connection type discriminator
+typedef enum {
+  NANO_CONN_WEBSOCKET,
+  NANO_CONN_HTTP_STREAM
+} nano_conn_type;
+
+// Connection state machine (prevents races between R and NNG threads)
+typedef enum {
+  CONN_STATE_OPEN,      // Connection active, can send/receive
+  CONN_STATE_CLOSING,   // Close initiated, waiting for cleanup
+  CONN_STATE_CLOSED     // Fully closed, safe to free
+} nano_conn_state;
+
+// Base connection structure - common fields for linked list and lifecycle
+typedef struct nano_conn_s {
+  nano_conn_type type;              // Connection type (for dispatch)
+  nng_aio *send_aio;                // For async send (both types)
+  nano_http_handler_info *handler;  // Back-reference to handler
+  struct nano_conn_s *next;         // Linked list
+  struct nano_conn_s *prev;         // Doubly-linked for O(1) removal
+  SEXP xptr;                        // R external pointer
+  int id;                           // Unique connection ID (server-wide)
+  nano_conn_state state;            // Connection state
+  int onclose_scheduled;            // Prevents duplicate on_close callbacks
+} nano_conn;
+
+// WebSocket connection (embeds base as first member)
+typedef struct nano_ws_conn_s {
+  nano_conn conn;                   // Base - MUST be first member
+  nng_stream *stream;               // WebSocket stream with automatic framing
+  nng_aio *recv_aio;                // For async receive loop
+} nano_ws_conn;
+
+// HTTP streaming connection (embeds base as first member)
+typedef struct nano_stream_conn_s {
+  nano_conn conn;                   // Base - MUST be first member
+  nng_http_conn *http;              // Hijacked HTTP connection
+  nng_http_req *req;                // Request pointer (valid because hijacked)
+  // Response state (set before first send)
+  char **resp_header_names;
+  char **resp_header_values;
+  int resp_header_count;
+  int resp_header_capacity;
+  int status_code;                  // Default 200
+  int headers_sent;                 // Whether HTTP headers written
+} nano_stream_conn;
+
+// HTTP handler info (links NNG handler to R callback)
+typedef struct nano_http_handler_info_s {
+  nng_http_handler *handler;        // NNG HTTP handler (NULL for WS)
+  SEXP callback;                    // HTTP callback or WS on_message or stream on_request
+  nano_http_server *server;         // Back-reference
+  // Long-lived connection handler fields:
+  nng_stream_listener *listener;    // WebSocket listener (NULL for HTTP stream)
+  nng_aio *accept_aio;              // Accept AIO for WS handler
+  nano_conn *conns;                 // Linked list of connections (WS or HTTP stream)
+  SEXP on_open;                     // R callback: WS open or stream request
+  SEXP on_close;                    // R callback: connection close
+  int textframes;                   // WS: text frame mode
+} nano_http_handler_info;
+
+typedef struct nano_http_request_s {
+  nng_aio *aio;                     // The HTTP AIO to complete
+  nng_http_req *req;                // The HTTP request (valid until aio finished)
+  SEXP callback;                    // R callback function
+  nano_http_server *server;         // Back-reference to server
+  nano_http_request *next;          // Linked list for tracking
+  nano_http_request *prev;
+  int cancelled;                    // Set when server stops (protected by server->mtx)
+} nano_http_request;
+
+typedef enum {
+  SERVER_CREATED,   // Server created but not started
+  SERVER_STARTED,   // Server running
+  SERVER_STOPPED    // Server stopped
+} nano_server_state;
+
+typedef struct nano_http_server_s {
+  nng_http_server *server;          // NNG HTTP server
+  nng_tls_config *tls;              // TLS configuration
+  nano_http_handler_info *handlers; // Array of handler info
+  int handler_count;                // Number of handlers
+  nano_http_request *pending_reqs;  // Linked list of pending HTTP requests
+  nng_mtx *mtx;                     // Mutex for thread safety
+  int conn_counter;                 // Server-wide unique connection ID counter
+  nano_server_state state;          // Server lifecycle state
+  SEXP xptr;                        // R external pointer for this server
+  SEXP prot;                        // Pairlist for GC protection of callbacks
+} nano_http_server;
+
+typedef struct ws_message_s {
+  nano_ws_conn *conn;
+  nng_msg *msg;
+} ws_message;
+
+#endif
 
 extern void (*eln2)(void (*)(void *), void *, double, int);
 
@@ -269,6 +381,8 @@ extern SEXP nano_StreamSymbol;
 extern SEXP nano_TlsSymbol;
 extern SEXP nano_UrlSymbol;
 extern SEXP nano_ValueSymbol;
+extern SEXP nano_HttpServerSymbol;
+extern SEXP nano_ConnSymbol;
 
 extern SEXP nano_aioFuncMsg;
 extern SEXP nano_aioFuncRes;
@@ -323,6 +437,7 @@ SEXP nano_raw_char(const unsigned char *, const size_t);
 void nano_serialize(nano_buf *, const SEXP, SEXP, int);
 SEXP nano_unserialize(unsigned char *, const size_t, SEXP);
 SEXP nano_decode(unsigned char *, const size_t, const uint8_t, SEXP);
+SEXP nano_url_with_port(nng_url *, int);
 void nano_encode(nano_buf *, const SEXP);
 int nano_encode_mode(const SEXP);
 int nano_matcharg(const SEXP);
@@ -333,6 +448,10 @@ SEXP nano_aio_http_status(SEXP);
 void pipe_cb_signal(nng_pipe, nng_pipe_ev, void *);
 void tls_finalizer(SEXP);
 
+void nano_load_later(void);
+SEXP nano_PreserveObject(const SEXP);
+void nano_ReleaseObject(SEXP);
+
 void nano_list_do(nano_list_op, nano_aio *);
 void nano_thread_shutdown(void);
 
@@ -362,11 +481,16 @@ SEXP rnng_cv_wait_safe(SEXP);
 SEXP rnng_dial(SEXP, SEXP, SEXP, SEXP, SEXP);
 SEXP rnng_dialer_close(SEXP);
 SEXP rnng_dialer_start(SEXP, SEXP);
+SEXP rnng_dispatcher_run(SEXP, SEXP, SEXP, SEXP, SEXP, SEXP, SEXP);
 SEXP rnng_eval_safe(SEXP);
 SEXP rnng_fini(void);
 SEXP rnng_fini_priors(void);
 SEXP rnng_get_opt(SEXP, SEXP);
 SEXP rnng_header_read(SEXP);
+SEXP rnng_http_server_close(SEXP);
+SEXP rnng_http_server_create(SEXP, SEXP, SEXP);
+SEXP rnng_http_server_start(SEXP);
+SEXP rnng_http_server_stop(SEXP);
 SEXP rnng_ip_addr(void);
 SEXP rnng_is_error_value(SEXP);
 SEXP rnng_is_nul_byte(SEXP);
@@ -410,10 +534,17 @@ SEXP rnng_tls_config(SEXP, SEXP, SEXP, SEXP);
 SEXP rnng_traverse_precious(void);
 SEXP rnng_unresolved(SEXP);
 SEXP rnng_unresolved2(SEXP);
+SEXP rnng_race_aio(SEXP, SEXP);
 SEXP rnng_url_parse(SEXP);
 SEXP rnng_version(void);
 SEXP rnng_wait_thread_create(SEXP);
 SEXP rnng_write_cert(SEXP, SEXP);
 SEXP rnng_write_stdout(SEXP);
+SEXP rnng_ws_close(SEXP);
+SEXP rnng_ws_send(SEXP, SEXP);
+SEXP rnng_conn_close(SEXP);
+SEXP rnng_stream_conn_send(SEXP, SEXP);
+SEXP rnng_stream_conn_set_header(SEXP, SEXP, SEXP);
+SEXP rnng_stream_conn_set_status(SEXP, SEXP);
 
 #endif
